--|----------------------------------------------------------------------------
--| This is a rewritten FORMAT class for the "SmallEiffel" compiler!
--| This class came originally from "Tower Technology Corporation". 
--| You need for this class additionally some external C routines, 
--| which are supplied in a related "format.c" file.
--|
--| Adapted $Date: 1996/06/26 11:13:19 $ by V. Kyriakides
--| Email: 1kyriaky@informatik.uni-hamburg.de
--| 
--|----------------------------------------------------------------------------
--| *** PLEASE NOTICE THE BELOW HEADER *** from Tower Technology Corporation!
--|----------------------------------------------------------------------------

--|----------------------------------------------------------------------------
--| TowerEiffel -- Copyright (c) 1993,1994,1995 Tower Technology Corporation. 
--| All Rights Reserved.
--| Please do not remove this header from this class text.
--|----------------------------------------------------------------------------
--
-- Formatted output and basic type conversion.
-- Interface conforms to the public interface for the SIG Eiffel/S
-- FORMAT class from their basic library, but extends the
-- functionality by allowing more control over formatting.
--
-- For i2s, r2s, d2s and s2s:
--   Follow one of these two rules for format strings:
--   1) Supply a complete "printf"-style format string
--      beginning with a `%' and ending with a type specifier.
--      Warning! Since `%' is the Eiffel string escape key, make sure
--      to escape it! Notice the pair of %'s in the following examples:
--
--      out_string := r2s( "%%5.2f", my_real );
--      out_string := i2s( "%%-+5d", my_int ) ;
--      out_string := i2s( "%%05x", my_int ) ;
--
--   2) Supply no leading `%' and no trailing type specifier
--      (i.e. d,f,g,e,s, etc.). In this case a reasonable
--      default type specifier is supplied for you. Examples:
--
--      out_string := r2s( "5.2", my_real ) ;
--      out_string := i2s( "-+5", my_int ) ;

indexing

   date: "$Date: 1996/06/26 11:13:19 $";
   revision: "$Revision: 1.1 $"


class FORMAT
   
feature {ANY}
   
   i2s (format : STRING; value : INTEGER) : STRING is
         -- String from an integer and a format string
      require
         format_not_void : format /= Void
         legal_format    : -- `format' must be a legal format string
      do
			use_format  := format ;
         fix_format_string( 'd' ) ;
         Result := ext_i2s(use_format.to_external, value)
      end ; -- i2s
   
   s2i (text : STRING) : INTEGER is
         -- Integer from a string representation.
      require
         text_not_void        : text /= Void
         valid_integer_string : -- `text' must represent an integer
      do
         -- this one happens to be defined in ISE's STRING class
         Result := text.to_integer
      end ; -- s2i
   
feature {ANY}
   
   r2s (format : STRING; value : REAL) : STRING is
         -- String from a real and a format string
      require
         format_not_void : format /= Void
         legal_format    : -- `format' must be a legal format string
      do
			use_format := format ;
         fix_format_string( 'f' ) ;
         Result := ext_r2s(use_format.to_external, value)
      end ; -- r2s
   
   s2r (text : STRING) : REAL is
         -- Real from a string representation.
      require
         text_not_void     : text /= Void
         valid_real_string : -- `text' must represent a real
      do
         -- this one happens to be defined in SmallEiffel's STRING class
			Result := text.to_real ;
      end ; -- s2r
   
feature {ANY}
   
   d2s (format : STRING; value : DOUBLE) : STRING is
         -- String from a double and a format string
      require
         format_not_void : format /= Void
         legal_format    : -- `format' must be a legal format string
      do
			use_format := format ;
         fix_format_string( 'g' ) ;
         Result := ext_d2s(use_format.to_external, value)
      end ; -- d2s
   
   s2d (text : STRING) : DOUBLE is
         -- Real from a string representation.
      require
         text_not_void     : text /= Void
         valid_real_string : -- `text' must represent a real
      do
         -- this one happens to be defined in SmallEiffel's STRING class
			Result := text.to_double ;
      end ; -- s2d
   
feature {ANY}
   
   c2s (c : CHARACTER) : STRING is
         -- String from a character
      do
         Result := c.out ;
      end ; -- c2s
   
   s2c (text : STRING) : CHARACTER is
         -- the first character of the provided string
      require
         long_enough : text /= Void and then text.count > 0
      do
         Result := text.item (1)
      end ; -- s2c
   
   s2s (format : STRING; text : STRING) : STRING is
         -- String from a string and a format string
      require
         format_not_void : format /= Void
         text_not_void   : text /= Void
      do
         use_format  := format ;
         fix_format_string( 's' ) ;
         Result := ext_s2s(use_format.to_external, text.to_external)
      end ; -- s2s
   
feature {ANY}
   
   b2s (b : BOOLEAN) : STRING is
         -- string from a boolean
      do
         if b then
	    		Result := "true" ;
         else
	    		Result := "false" ;
         end
      end ; -- b2s
   
   s2b (text : STRING) : BOOLEAN is
         -- boolean from a string
      require
         text_not_void        : text /= Void
         valid_boolean_string : -- `text' must represent a boolean
      do
         if text = "true" or text = "TRUE" then
            Result := TRUE
         end
      end ; -- s2b

feature {NONE}
   
   use_format : STRING ;
   
   fix_format_string( c : CHARACTER ) is
         -- make sure the format is reasonable and has
         -- `c' as the final character
      require
         use_format_count_ok : use_format.count /= 1
      do
         use_format.left_adjust ;
         use_format := use_format.right_trim(use_format) ;
         if use_format.count > 1 then
            if use_format @ 1 /= '%%' then
               use_format.precede( '%%' )
               use_format.extend( c )
            end
         else
            use_format.extend( '%%' ) ;
            use_format.extend( c ) ;
         end
      end ; -- fix_format_string
   
	
   ext_i2s( format : POINTER ; value : INTEGER ) : STRING is
         -- return a string from an integer
      external "C"
      end ; -- ext_i2s
      
	
   ext_r2s( format : POINTER ; value : REAL ) : STRING is
         -- return a string from a real
      external "C"
      end ; -- ext_i2s
      

   ext_d2s( format : POINTER ; value : DOUBLE ) : STRING is
         -- return a string from a double
      external "C"
      end ; -- ext_i2s
   

   ext_s2s( format, value : POINTER ) : STRING is
         -- return a string from a c string
      external "C"
      end ; -- ext_i2s
   
   
end -- class FORMAT

