-- Part of SmallEiffel -- Read DISCLAIMER file -- Copyright (C) 
-- Dominique COLNET and Suzanne COLLIN -- colnet@loria.fr
--
class ARRAY[E]
-- 
-- General purpose arrays.
--

inherit COLLECTION[E];

creation make, with_capacity

feature
   
   capacity: INTEGER;
	 -- Internal storage capacity in number of item.
   
   lower: INTEGER;     
	 -- Lower index bound.
   
   upper: INTEGER;
	 -- Upper index bound.
   
feature {NONE}
   
   storage: POINTER;
	 -- Internal access to storage location.
   
feature -- Creation and Modification :
   
   make(minindex, maxindex: INTEGER) is
	 -- Make array with range [minindex .. maxindex]. 
	 -- When maxindex = minindex - 1, the array is empty.
      require
	 minindex -1 <= maxindex
      local
	 mem: MEMORY;
      do
	 lower := minindex;
	 upper := maxindex;
	 if storage.is_not_void then
	    mem.free(storage);
	 end;
	 capacity := maxindex - minindex + 1;
	 if capacity > 0 then
	    if capacity < 16 then capacity := 16 end;
	    storage := malloc(capacity);
	    clear_all;
	 end;
      ensure
	 lower = minindex;
	 upper = maxindex;
	 all_cleared;
      end;

   with_capacity(needed_capacity, low: INTEGER) is
	 -- Create an empty array with `capacity' initialized
	 -- to `needed_capacity' and `lower' set to `low'.
      require
	 needed_capacity > 0
      do
	 make(low,low + needed_capacity - 1);
	 upper := low - 1;
      ensure
	 empty;
	 needed_capacity <= capacity;
	 lower = low
      end;

feature -- Modification :
   
   resize(minindex, maxindex: INTEGER) is
	 -- Resize the array. No elements will be lost in the 
	 -- intersection of [old lower .. old upper] and 
	 -- [minindex .. maxindex].
	 -- New positions are initialized with appropriate 
	 -- default values.
      require
	 minindex <= maxindex
      local
	 other: like Current;
	 i, up: INTEGER;
	 mem: MEMORY;
      do
	 from
	    !!other.make(minindex,maxindex);
	    i := lower.max(other.lower);
	    up := upper.min(other.upper);
	 until
	    i > up
	 loop
	    other.put(item(i),i);
	    i := i + 1;
	 end;
	 if storage.is_not_void then
	    mem.free(storage);
	 end;
	 standard_copy(other);
	 mem.free(other.to_pointer);
      end;

feature -- No modification :

   item, infix "@" (index: INTEGER): E is
      external "CSE"
      end;
   
feature -- Modification :
   
   put(element: E; index: INTEGER) is
      external "CSE"
      end;

   force(element: E; index: INTEGER) is
      do
	 if upper < index then
	    resize(lower,index);
	 elseif index < lower then
	    resize(index,upper);
	 end;
	 put(element,index);
      end;

   copy(other: like Current) is
	 -- Copy `other' onto Current.
      local
	 i: INTEGER;
      do
	 upper := lower - 1;
	 if capacity = 0 then
	    make(other.lower,other.upper);
	 else
	    resize(other.lower,other.upper);
	 end;
	 from
	    i := lower;
	 until
	    i > upper
	 loop
	    put(other.item(i),i);
	    i := i + 1;
	 end;
      end;
   
   remove(index: INTEGER) is
      local
	 i: INTEGER;
      do
	 from  
	    i := index;
	 until
	    i + 1 > upper
	 loop
	    put(item(i + 1),i);
	    i := i + 1;
	 end;
	 upper := upper - 1;
      end;
   
   remove_first is
      local
	 i: INTEGER;
      do
	 from
	    i := lower;
	 until
	    i = upper
	 loop
	    put(item(i+1),i);
	    i := i + 1;
	 end;
	 lower := lower + 1;
      ensure then
	 upper = old upper;
      end;
   
   remove_last is
      require
	 not empty;
      do
	 upper := upper - 1;
      end;
   
   clear is
      do
	 upper := lower - 1;
      ensure
	 empty;
      end;

feature -- Adding Elements :
   
   add_last(elt: E) is
      do
	 if storage = Void then end;
	 if capacity < count + 1 then
	    if capacity = 0 then
	       capacity := 16;
	       storage := malloc(capacity);
	    else
	       capacity := capacity + 16;
	       storage := realloc(storage,capacity);
	    end;
	 end;
	 upper := upper + 1;
	 put(elt,upper);
      ensure
	 last = elt;
	 count = 1 + old count
      end;
   
   sub_array(low, up: INTEGER): ARRAY[E] is
      require
	 valid_index(low);
	 valid_index(up);
	 low <= up
      local
	 i: INTEGER;
      do
	 from  
	    !!Result.make(low,up);
	    i := low;
	 until
	    i > up
	 loop
	    Result.put(item(i),i);
	    i := i + 1;
	 end;
      ensure
	 Result.lower = low;
	 Result.upper = up;
      end;

feature -- Interfacing with C :
   
   to_external: POINTER is
	 -- Gives C access into the internal `storage' of the ARRAY.
	 -- Result is pointing the element at index `lower'.
	 -- 
	 -- NOTE: do not free/realloc the Result. Resizing of the array 
	 --       can makes this pointer invalid. 
      require
	 not empty
      do
	 Result := storage;
      ensure
	 Result.is_not_void
      end;

feature {NONE}

   malloc(size: INTEGER): POINTER is
      require
	 size > 0
      local
	 x: like item;
	 mem: MEMORY;
      do
	 if x.is_expanded_type then
	    Result := mem.malloc(size * x.object_size);
	 else
	    Result := mem.malloc(size * mem.pointer_size);
	 end;
      end;

   realloc(pointer: POINTER; size: INTEGER): POINTER is
      require
	 size > 0
      local
	 x: like item;
	 mem: MEMORY;
      do
	 if x.is_expanded_type then
	    Result := mem.realloc(pointer,size * x.object_size);
	 else
	    Result := mem.realloc(pointer,size * mem.pointer_size);
	 end;
      end;

invariant
   
   capacity >= (upper - lower + 1);

   capacity > 0 implies storage.is_not_void;

end -- ARRAY[E]
