expanded class BIT_N
--
-- Indexed Bit sequences of length N. This class is a template,
-- not a real class; to obtain a meaningful class, replace N 
-- with a positive integer throughout.
-- 
-- An INTEGER index can be used to access each bit of the sequence.
-- The leftmost bit has index 1 and the rightmost bit has index N.
--
-- Note : corresponding C mapping is done using a sequence of C char
--        inside a wrapping C struct. The name of the member is 
--        always `storage'. Assume there is enougth space for 8 bits
--        in one char. Examples :
--           Eiffel declaration | C declaration
--           --------------------------------------------------
--           BIT 1              | struct {char storage[1];};
--           BIT 2              | struct {char storage[1];};
--           BIT 8              | struct {char storage[1];};
--           BIT 9              | struct {char storage[2];};
--

inherit ANY;

feature {NONE}

   storage: POINTER;
	 -- The beginning of the storage zone (first byte
	 -- contains the beginning of the sequence).

feature -- Basic Accessing :
   
   count: INTEGER is
	 -- Number of bits in the sequence (the value of N).
      external "CSE"
      end;
   
   item(idx: INTEGER): BOOLEAN is
	 -- True if i-th bit is 1, false otherwise.
      require
	 inside_bounds: 1 <= idx and then idx <= count
      do
	 c_inline_c(
	 "{unsigned char c=((C->storage)[(--a1)/8]);%N%
         % c=c>>(7-(a1%%8));%N%
	 % R=((c%%2)?1:0);}");
      end;
   
   put(value: BOOLEAN; idx: INTEGER) is
	 -- Set bit `idx' to 1 if value is true, 0 otherwise.
      require
	 inside_bounds: 1 <= idx and then idx <= count
      do
	 if value then
	    put_1(idx);
	 else
	    put_0(idx);
	 end;
      ensure
	 value = item(idx)
      end;

   put_1(idx: INTEGER) is
	 -- Set bit `idx' to 1.
      do
	 c_inline_c(
	 "{unsigned char*p=((unsigned char*)(C->storage))+((--a1)/8);%N%
	 % unsigned char m=(1<<(7-(a1%%8)));%N%
	 % *p|=m;}");
      end;

   put_0(idx: INTEGER) is
	 -- Set bit `idx' to 0.
      do
	 c_inline_c(
	 "{unsigned char*p=((unsigned char*)(C->storage))+((--a1)/8);%N%
	 % unsigned char m=(1<<(7-(a1%%8)));%N%
	 % *p&=(~m);}");
      end;

feature -- Rotating and shifting :

   infix "^" (s: INTEGER): like Current is 
	 -- Retur a new sequence shifted by `s' positions.
         -- Positive `s' shifts right, negative left.
         -- Incoming bits are 0. Falling bits are lost.
      do
         Result := Current;
	 if s > 0 then
            Result.shift_right(s);
	 else
	    Result.shift_left(s);
	 end;
      end;

   shift_right(s: INTEGER) is
	 -- Sequence shifted right by `s' positions.
	 -- Incoming bits are 0. Falling bits are lost.
      require
         s >= 0
      local
         i1, i2: INTEGER;
      do
	 from
            i1 := count - s;
	    i2 := count;
	 until
            i1 = 0
	 loop
            put(item(i1),i2);
	    i1 := i1 - 1;
	    i2 := i2 - 1;
	 end;
	 from
            i1 := s;
	 until
            i1 = 0
	 loop
            put_0(i1);
	    i1 := i1 - 1;
	 end;
      end;

   shift_left(s: INTEGER) is
	 -- Sequence shifted left by `s' positions.
	 -- Incoming bits are 0. Falling bits are lost.
      require
         s >= 0
      local
         i1, i2: INTEGER;
      do
	 from
            i1 := 1;
	    i2 := 1 + s;
	 until
            i2 > count
	 loop
            put(item(i2),i1);
	    i1 := i1 + 1;
	    i2 := i2 + 1;
	 end;
	 from
	 until
            i1 > count
	 loop
            put_0(i1);
	    i1 := i1 + 1;
	 end;
      end;

   infix "#" (s: INTEGER): like Current is 
	 -- Sequence rotated by `s' positions (positive right,
	 -- negative left).
      do
	 Result := Current;
	 if s > 0 then
            Result.rotate_right(s);
	 else
	    Result.rotate_left(s);
	 end;
      end;

   rotate_right(s: INTEGER) is 
	 -- Sequence rotated right by `s' positions.
      require
         s >= 0
      local
         i: INTEGER;
      do
         from
            i := s;
	 until
            i = 0
	 loop
            rotate_right_1;
            i := i - 1;
	 end;
      end;

   rotate_right_1 is 
	 -- Sequence rotated right by 1 position.
      local
         bit: BOOLEAN;
      do
         bit := item(count);
         shift_right(1);
         put(bit,1);
      end;

   rotate_left(s: INTEGER) is 
	 -- Sequence rotated left by `s' positions.
      require
         s >= 0
      local
         i: INTEGER;
      do
         from
            i := s;
	 until
            i = 0
	 loop
            rotate_left_1;
            i := i - 1;
	 end;
      end;

   rotate_left_1 is 
	 -- Sequence rotated left by 1 position.
      local
         bit: BOOLEAN;
      do
         bit := item(1);
         shift_left(1);
         put(bit,count);
      end;

feature -- Bitwise Logical Operators :

   infix "and" (other: like Current): like Current is
	 -- Bitwise `and' of Current with `other'
      do
	 not_yet_implemented;
      end;

   infix "or" (other: like Current): like Current is
	 -- Bitwise `or' of Current with `other'
      do
	 not_yet_implemented;
      end;

   infix "xor" (other : like Current) : like Current is
	 -- Bitwise `xor' of Current with `other'
      do
	 not_yet_implemented;
      end;

   infix "implies" (other: like Current): like Current is
	 -- Bitwise implication of Current with `other'
      do
	 not_yet_implemented;
      end;

   prefix "not": like Current is
	 -- Bitwise `not' of Current.
      do
	 not_yet_implemented;
      end;

feature -- Conversions :

   to_string: STRING is
	 -- String representation of bit sequence
	 -- A zero bit is mapped to '0', a one bit 
	 -- to '1'. item (1) is rightmost character
	 -- in string.
	 --
	 -- Note: see `append_in' to save memory.
      do
	 tmp_string.clear;
	 append_in(tmp_string);
	 Result := tmp_string.twin;
      ensure then
	 Result.count = count
      end;
   
   append_in(str: STRING) is
      local
	 i: INTEGER;
      do
	 from  
	    i := 1;
	 until
	    i > count
	 loop
	    if item(i) then
	       str.extend('1');
	    else
	       str.extend('0');
	    end;
	    i := i + 1;
	 end;
      end;
   
   from_string(s: STRING) is
	 -- Fill sequence from string `s'. 
	 -- Assume `s' contains `0' and `1' characters only.
      require
	 count = s.count
      local
	 i: INTEGER;
      do
	 from
	    i := s.count;
	 until
	    i = 0
	 loop
	    if s.item(i) = '1' then
	       put(true,i);
	    else
	       check
		  s.item(i) = '0'
	       end;
	       put(false,i);
	    end;
	    i := i - 1;
	 end;
      ensure
	 to_string.is_equal(s)
      end; 

feature -- Other features :

   clear_all is
         -- Set all bits to 0.
      local
         i: INTEGER;
      do
         from
            i := count;
         until
            i = 0
         loop
            put_0(i);
            i := i - 1;
         end;
      end;

   set_all is
         -- Set all bits to 1.
      local
         i: INTEGER;
      do
         from
            i := count;
         until
            i = 0
         loop
            put_1(i);
            i := i - 1;
         end;
      end;

feature {NONE}

   tmp_string: STRING is
      once
         !!Result.make(128);
      end;

end -- BIT_N
