* apollo.asm - Apollo specific module for DEC's Scheme->C
*
* This file implements the assembly language part of the Apollo port,
* specifically for the DN3000 and DN4000 series, that is, depending on
* the M68020 CPU and M68881 FPP.
*
* Included are all the necessary math routines to catch integer overflow.
*
* This file is written for PIC (Position Independent Code), to build
* a shared library.
*
* Ray Lischner (uunet!mntgfx!lisch)
* 26 April 1990

	module	sc_apollo
	cpu	68020,68881
	sri	68020

	entry	sc_setregs
	entry	sc_regs
	entry	sc_iplus
	entry	sc_idifference
	entry	sc_inegate
	entry	sc_itimes

	data

* set up jump tables for calling PIC routines
data_start	equ	*

sc_iplus	lea	data_start,a0
		jmp.l	sc$iplus
sc_idifference	lea	data_start,a0
		jmp.l	sc$idifference
sc_inegate	lea	data_start,a0
		jmp.l	sc$inegate
sc_itimes	lea	data_start,a0
		jmp.l	sc$itimes

* set up transfer address for external PIC routines
	extern	sc_makefloat64
sc$makefloat64	ac	sc_makefloat64

	text

***********************************************************************
* void sc_setregs(int* a6, int* a7)
* Apollo's longjmp() checks to see if the jump is backwards in the stack.
* If not, it assumes that something is wrong and ungracefully terminates
* the program.  Since we don't want this to happen, we need to fake
* out Domain/OS.  This is done by setting the stack pointer (a7) and
* frame pointer (a6) to the destination frame, thus circumventing
* longjmp's checks.
*
* To accomplish this takes some clever tricks.  First, we need to know
* how the stack is layed out:
*
* (lower addresses)
*       +----------------------------+
*    A7 |  local storage ...         |
*       +----------------------------+
*    A6 | link to previous frame     |
*       +----------------------------+
*       | return address             |
*       +----------------------------+
*       | arguments pushed by caller |
*       +----------------------------+
* (higher addresses)
* Note that we are ignoring floating point control blocks.
*
* The caller pushes the desired values for A7 and A6.  On entry to sc_setregs(),
* A6 points to the caller's frame, and A7 points to the return address.
* We can retrieve the caller's arguments by dereferencing a7: the second
* argument is in 8(a7), and the first is in 4(a7).  We can just copy
* them into the registers we want, but first we need to save the return
* address before we lose the pointer to it.  It is saved in A0, at the
* same time we load A6 and A7.  Clever, isn't it?  After getting the new
* register values, we know that the caller will try to pop the
* arguments off the stack by adding 8 to A7.  We circumvent this by
* subtracting 8 now.

sc_setregs	procedure "sc_setregs",nocode
	movem.l	(a7),a0/a6-a7
	subq.l	#8,a7
	jmp	(a0)

***********************************************************************
* void sc_regs(int regs[12])
* sc_regs returns the values of a1-a4, d0-d7 in the caller supplied buffer.
* These are the "callee" save registers that need to be examined during
* garbage collection.

sc_regs	procedure	"sc_regs",#-4
	move.l	8(a6),a0		* a0 := &regs[0]
	movem.l	d0-d7/a1-a4,(a0)	* save the interesting registers
	return	sc_regs

***********************************************************************
* The following routines are for doing arithmetic on tagged numbers.
* The input arguments are tagged integers, that is, integers shifted
* left by two bits.  (Except for sc_itimes, where only the second
* argument, b, is shifted.)  This makes it easier to check for overflow,
* but we must unshift the values before calling sc_makefloat64().
*
* When the result of any operation overflows, the operands are converted
* to floating point, and the operation is repeated.  The floating point
* result is then passed to sc_makefloat64() to produce a float object
* to return.

* int sc_iplus(int a, int b)
*	returns the integer sum, a + b, where a and b are the two 
*	integer arguments, unless integer overflow occurs, then returns
*	(unsigned int) sc_makefloat64( (double)a + (double)b ) instead.

sc$iplus	procedure	"sc_iplus",#0,a5
	move.l	a0,a5
* add the arguments
	move.l	8(a6),d0
	move.l	12(a6),d1
	add.l	d1,d0
* if the operation overflows, we know to use floating point
	bvc	1$
	
* otherwise, convert to floating point and add
	move.l	8(a6),d0
	asr.l	#2,d0
	fmove.l	d0,fp0
* note that d1 still contains "b"
	asr.l	#2,d1
	fmove.l	d1,fp1
	fadd	fp1,fp0
* pass the floating point sum to sc_makefloat64
	fmove.d	fp0,-(sp)
	move.l	sc$makefloat64,a0
	jsr	(a0)
	addq.l	#8,sp

1$	return	sc$iplus


* int sc_idifference(int a, int b)
*	returns integer difference, a - b, where a and b are the two 
*	integer arguments, unless integer overflow occurs, then returns
*	(unsigned int) sc_makefloat64( (double)a - (double)b ) instead.
*

sc$idifference	procedure	"sc_idifference",#0,a5
	move.l	a0,a5
* subtract the arguments
	move.l	8(a6),d0
	move.l	12(a6),d1
	sub.l	d1,d0
* if the operation overflows, we know to use floating point
	bvc	1$

* otherwise, convert to floating point and subtract
	move.l	8(a6),d0
	asr.l	#2,d0
	fmove.l	d0,fp0
* note that d1 still contains "b"
	asr.l	#2,d1
	fmove.l	d1,fp1
	fsub	fp1,fp0
* pass the floating point sum to sc_makefloat64
	fmove.d	fp0,-(sp)
	move.l	sc$makefloat64,a0
	jsr	(a0)
	addq.l	#8,sp

1$	return	sc$idifference

* int sc_inegate(int a)
*	returns integer negation, -a, where a is the integer
*	argument, unless integer overflow occurs, then returns
*	(unsigned int) sc_makefloat64( -(double)a) instead.
*

sc$inegate procedure	"sc_inegate",#0,a5
	move.l	a0,a5
* negate the argument
	move.l	8(a6),d0
	move.l	d0,d1
	neg.l	d0
* if the operation overflows, we know to use floating point
	bvc	1$

* otherwise, convert to floating point and negate
	asr.l	#2,d1
	fmove.l	d1,fp1
	fneg	fp1,fp0
* pass the floating point sum to sc_makefloat64
	fmove.d	fp0,-(sp)
	move.l	sc$makefloat64,a0
	jsr	(a0)
	addq.l	#8,sp

1$	return	sc$inegate
	
* sc_itimes(int a, int b)
*	returns integer procuct, a * b, where a and b are the two 
*	integer arguments, unless integer overflow occurs, then returns
*	(unsigned int) sc_makefloat64( (double)a * (double)b ) instead.
* Unlike the previous arithmetic functions, only "b" has been shifted.

sc$itimes	procedure	"sc_itimes",#0,a5
	move.l	a0,a5
* multiply the arguments
	move.l	8(a6),d0
	move.l	12(a6),d1
	muls.l	d1,d0

* if the operation overflows, we know to use floating point
	bvc	1$

* otherwise, convert to floating point and multiply
	fmove.l	8(a6),fp0
* note that d1 still contains "b"
	asr.l	#2,d1
	fmove.l	d1,fp1
	fmul	fp1,fp0
* pass the floating point sum to sc_makefloat64
	fmove.d	fp0,-(sp)
	move.l	sc$makefloat64,a0
	jsr	(a0)
	addq.l	#8,sp

1$	return	sc$itimes

	end
