* prism.asm - Apollo Prism (DN10000) specific module for DEC's Scheme->C
*
* This file implements the assembly language part of the Prism port,
* specifically for the DN10000.
*
* Included are all the necessary math routines to catch integer overflow.
*
* NOTE: Don't even try to read this file if you do not understand
* how an Apollo Prism (also called an AT, for Advanced Technology;
* perhaps Apollo thinks the Prism is as good as an IBM PC AT :-) works.
* I have tried to optimize the parallel operations, such as branch and
* call shadows, and combining integer and floating point operations.
* (The former are common; the latter are rare in this file.)
*
* The sematics of b.sa are completely different from b.sn, and the
* subtle differences are too lengthy to discuss here.  Read the
* various Apollo manuals, such as the AT Assembler Reference and
* the AT Technical Reference.
*
* Apollo's setjmp/longjmp do not permit jumps to random locations in the
* stack, so we must write our own.  On the DN3000 (M68K), we can get away
* with simply altering the stack and frame pointers (A6 and A7) before
* calling longjmp, but on the Prism this does not work because longjmp
* only jumps to a valid stack frame.  I tried modifying call/cc to
* restore the stack and registers before calling longjmp(), but this
* does not work because it changes the data base register, which messes
* up the call to longjmp.  The simplest solution is to reimplemen]t
* setjmp and longjmp.
*
* Another reason to write our own setjmp/longjmp is to make sure all
* the registers are saved properly.  The standard jmp_buf does not have
* enough room to save all the needed registers.
*
* Ray Lischner (uunet!mntgfx!lisch)
* 1 May 1990

	module	sc_prism

	export.f	prism_stack_frame
	export.p	sc_longjmp
	export.f	sc_setjmp
	export.p	sc_regs
	export.f	sc_iplus
	export.f	sc_idifference
	export.f	sc_inegate
	export.f	sc_itimes

	import.f	sc_makefloat64
	import.f	sigblock
	import.f	sigsetmask

	data

* set up ECBs for all the functions that need one
data_frame	equ	*

sc_setjmp	procedure	ok
	.0 = sc$setjmp		; get the relocatable address of the routine
	.1 = .sf		; save the old stack frame
	.2 = #data_frame	; get the relocatable data frame
	b.sa [.0]		; branch to the real routine
	[--.sf,#16] = .1	; push the old .SF onto the stack

sc_longjmp	procedure	ok
	.0 = sc$longjmp		; get the relocatable address of the routine
	.1 = .sf		; save the old stack frame
	.2 = #data_frame	; get the relocatable data frame
	b.sa [.0]		; branch to the real routine
	[--.sf,#16] = .1	; push the old .SF onto the stack

sc_idifference	procedure	ok
	.0 = sc$idifference	; get the relocatable address of the routine
	.1 = .sf		; save the old stack frame
	.2 = #data_frame	; get the relocatable data frame
	b.sa [.0]		; branch to the real routine
	[--.sf,#16] = .1	; push the old .SF onto the stack

sc_inegate	procedure	ok
	.0 = sc$inegate		; get the relocatable address of the routine
	.1 = .sf		; save the old stack frame
	.2 = #data_frame	; get the relocatable data frame
	b.sa [.0]		; branch to the real routine
	[--.sf,#16] = .1	; push the old .SF onto the stack

sc_iplus	procedure	ok
	.0 = sc$iplus		; get the relocatable address of the routine
	.1 = .sf		; save the old stack frame
	.2 = #data_frame	; get the relocatable data frame
	b.sa [.0]		; branch to the real routine
	[--.sf,#16] = .1	; push the old .SF onto the stack

sc_itimes	procedure	ok
	.0 = sc$itimes		; get the relocatable address of the routine
	.1 = .sf		; save the old stack frame
	.2 = #data_frame	; get the relocatable data frame
	b.sa [.0]		; branch to the real routine
	[--.sf,#16] = .1	; push the old .SF onto the stack

* jump table for the ECBs
sc$setjmp	data.l	sc$$setjmp
sc$longjmp	data.l	sc$$longjmp
sc$idifference	data.l	sc$$idifference
sc$inegate	data.l	sc$$inegate
sc$iplus	data.l	sc$$iplus
sc$itimes	data.l	sc$$itimes

* relocation table for the external functions
sc$makefloat64	data.l	sc_makefloat64
sig$setmask	data.l	sigsetmask
sig$block	data.l	sigblock
	proc

***********************************************************************
* int prism_stack_frame(void)
* Return the caller's stack frame pointer.  See the STACKPTR macro
* in heap.h for how this is called.

prism_stack_frame	procedure ok
	b.sa [.return]
	.0 = .sf


***********************************************************************
* int sc_setjmp(jmp_buf buf)
* Save the current signal mask, processor status words, and preserved
* registers in the caller-supplied buffer, and return zero.

sc$$setjmp	procedure name="sc_setjmp",return=save,stack=(),save=1$
	[.sf,4] = .4
	[.sf,8] = .return
	[.sf,12] = .10
	.10 = .2
	using .10, data_frame
1$	.0 = .ipsw		; and the processor status words
	[.4++] = .0
	.0 = .fppsw
	[.4++] = .0
	.0 = [.sf,12]	; old value of .10
	[.4++] = .0
	[.4++] = .11
	[.4++] = .12
	[.4++] = .13
	[.4++] = .14
	[.4++] = .15
	[.4++] = .16
	[.4++] = .17
	[.4++] = .18
	[.4++] = .19
	[.4++] = .20
	[.4++] = .21
	[.4++] = .return
	.0 = [.sf]
	[.4++] = .0

	.3 = sig$block		; and the current signal mask
	.return = call.sa [.3]	; sigblock(0)
	.4 = .null
	.4 = [.sf,4]
	[.4++] = .0

	.10 = [.sf,12]		; restore the saved registers
	.return = [.sf,8]
	.0 = .null
	b.sa [.return]		; return(0)
	.sf = [.sf]

	drop .10
	endp

* void longjmp(jmp_buf buf, int rtn)
* Jump to the location saved by a previous call to setjmp(), such that
* it looks to the caller of setjmp() as though setjmp returned "rtn".
* If "rtn" is zero, one is returned.

sc$$longjmp	procedure name="sc_longjmp",return=save,stack=(),save=1$
	[.sf,8] = .return
	using .2, data_frame
1$	.cc = .5
	.0 = [.4++]
	bnz.sf 2$	; make sure the return value is non-zero
	.5 = #1
2$	.ipsw = .0
	.0 = [.4++]
	.fppsw = .0
	.10 = [.4++]
	.11 = [.4++]
	.12 = [.4++]
	.13 = [.4++]
	.14 = [.4++]
	.15 = [.4++]
	.16 = [.4++]
	.17 = [.4++]
	.18 = [.4++]
	.19 = [.4++]
	.20 = [.4++]
	.21 = [.4++]
	.0 = [.4++]
	[.sf,8] = .0		; save the return PC
	.0 = [.4++]
	[.sf] = .0		; save .sf
	[.sf,4] = .5		; save return value

	.3 = sig$setmask	; restore the signal mask
	.return = call.sa [.3]
	.4 = [.4]

	.0 = [.sf,4]		; return the user-supplied "rtn"
	.return = [.sf,8]
	b.sa [.return]
	.sf = [.sf]
	endp

***********************************************************************
* void sc_regs(int regs[12])
* sc_regs stores the values of .10 - .21 in the caller supplied buffer.
* These are the "callee" save registers that need to be examined during
* garbage collection.

sc_regs	procedure ok
	[.4++] = .10
	[.4++] = .11
	[.4++] = .12
	[.4++] = .13
	[.4++] = .14
	[.4++] = .15
	[.4++] = .16
	[.4++] = .17
	[.4++] = .18
	[.4++] = .19
	[.4++] = .20
	b.sa [.return]
	[.4] = .21
	endp

***********************************************************************
* The following routines are for doing arithmetic on tagged numbers.
* The input arguments are tagged integers, that is, integers shifted
* left by two bits.  (Except for sc_itimes, where only the second
* argument, b, is shifted.)  This makes it easier to check for overflow,
* but we must unshift the values before calling sc_makefloat64().
*
* When the result of any operation overflows, the operands are converted
* to floating point, and the operation is repeated.  The floating point
* result is then passed to sc_makefloat64() to produce a float object
* to return.


* int sc_iplus(int a, int b)
*	returns the integer sum, a + b, where a and b are the two 
*	integer arguments, unless integer overflow occurs, then returns
*	(unsigned int) sc_makefloat64( (double)a + (double)b ) instead.

sc$$iplus	procedure	name="sc_iplus",return=save,stack=(),save=1$
	[.sf,8] = .return
* add the arguments
1$	.0.cc = .4 + .5		; try adding the arguments as integers
	.4 = .4 SHRA #2		; wait 1 cycle until CCs set
	bnv.sf 2$		; return if the integer operation worked
	.5 = .5 SHRA #2		; otherwise keep working
	.fs0.i = .5		; convert the integers to floating point
	.fs1.i = .4
	.fd8 = float(.fs1.i)
	.fd0 = float(.fs0.i)
* get ready to call makefloat64, while adding the operands
	.3 = sc$makefloat64, .fd8 += .fd0
	.return = call.sn [.3]	; call sc_makefloat64()
	nop
	.return = [.sf,8]	; pop the return PC

2$	b.sa [.return]		; return
	.sf = [.sf]		; restore the old .SF
	endp


* int sc_idifference(int a, int b)
*	returns integer difference, a - b, where a and b are the two 
*	integer arguments, unless integer overflow occurs, then returns
*	(unsigned int) sc_makefloat64( (double)a - (double)b ) instead.

sc$$idifference	procedure	ok,name="sc_idifference"
	[.sf,8] = .return
* subtract the arguments
1$	.0.cc = .4 - .5		; try subtracting the arguments as integers
	.4 = .4 SHRA #2		; wait 1 cycle until CCs set
	bnv.sf 2$		; return if the integer operation worked
	.5 = .5 SHRA #2		; otherwise keep working
	.fs0.i = .5		; convert the integers to floating point
	.fs1.i = .4
	.fd8 = float(.fs1.i)
	.fd0 = float(.fs0.i)
* get ready to call makefloat64, while subtracting the operands
	.3 = sc$makefloat64, .fd8 -= .fd0
	.return = call.sn [.3]	; call sc_makefloat64()
	nop
	.return = [.sf,8]	; pop the return PC

2$	b.sa [.return]		; return
	.sf = [.sf]		; restore the old .SF
	endp

* int sc_inegate(int a)
*	returns integer negation, -a, where a is the integer
*	argument, unless integer overflow occurs, then returns
*	(unsigned int) sc_makefloat64( -(double)a) instead.

sc$$inegate procedure	ok,name="sc_inegate"
	[.sf,8] = .return
* negate the argument
1$	.0.cc = -.4		; try negating the argument as an integer
	.4 = .4 SHRA #2		; wait 1 cycle until CCs set
	bnv.sf 2$		; return if the integer operation worked
	.fs0.i = .4		; otherwise keep working
	.fd8 = float(.fs0.i)	; convert the argument to floating point
* get ready to call makefloat64, while negating the argument
	.3 = sc$makefloat64, .fd8 = -.fd8
	.return = call.sn [.3]	; call sc_makefloat64()
	nop
	.return = [.sf,8]	; pop the return PC

2$	b.sa [.return]		; return
	.sf = [.sf]		; restore the old .SF
	endp
	
* sc_itimes(int a, int b)
*	returns integer procuct, a * b, where a and b are the two 
*	integer arguments, unless integer overflow occurs, then returns
*	(unsigned int) sc_makefloat64( (double)a * (double)b ) instead.
*	Unlike the previous arithmetic functions, only "b" has been shifted.
*
* 	This is a pain on a Prism because we need to use the floating
*	point unit for the integer multiply, and that means we cannot
*	set the integer condition codes.  Instead, we do a normal
*	floating point multiply and explicitly check the result to see
*	if it fits into an integer.  If not, we divide by 4 to get the
*	true result.  Note that this does not affect the precision
*	of the result.

sc$$itimes	procedure	name="sc_itimes",return=save,stack=(),save=1$
	[.sf,8] = .return
1$	.fs0.i = .4		; load floating point registers for the
	.fs1.i = .5		; multiplication
	.fd8 = float(.fs0.i)
	.fd2 = float(.fs1.i)
* do the multiply; at the same time, load sc_makefloat64's address, to
* get ready for calling it, in case the multiply overflows
	.3 = sc$makefloat64, .fd8 *= .fd2

* The floating point constants do not change, so we can put them in
* the shared text segment.  Change the address base to .PC, so we
* use PC-relative addressing.
	drop .2

* look for overflow by comparing with the maximum allowable integer
	.fd2 = maxint		; get maxint
	.fcc = .fd8 ? .fd2
	bfgt.sf 2$		; see if the result fits into an integer
	.fd2 = minint.fd
	.fcc = .fd8 ? .fd2
	bflt.sf 2$

	.fs0.i = round(.fd8)	; yes, so convert it to an integer
	b.sa 3$			; and return
	.0 = .fs0.i

2$	.fd0 = four
	.fd8 /= .fd0		; get the real floating point value
	.return = call.sn [.3]	; call sc_makefloat64()
	nop
	.return = [.sf,8]	; pop the return PC

3$	b.sa [.return]		; return
	.sf = [.sf]		; restore the old .SF

* constant value for the division, above
four	data.fd	4.0

* maximum and minimum possible integer, for comparison, above
maxint	data.fd  2147483644.0

* The assembler seems to ignore the sign of a floating point constant.
* A Prism uses IEEE format, so the smallest possible integer
* is pretty easy to write in hexadecimal.
*minint	data.fd -2147483648.0
minint	data.q	h'C1E00000, h'00000000
	endp

	end
