"File Extras.st
======================================================================
|
| Copyright (C) 1995 Free Software Foundation, Inc.
| Written by Brad Diller.
|
| This file is part of GNU Smalltalk.
|
| GNU Smalltalk is free software; you can redistribute it and/or modify it
| under the terms of the GNU General Public License as published by the Free
| Software Foundation; either version 1, or (at your option) any later version.
| 
| GNU Smalltalk is distributed in the hope that it will be useful, but WITHOUT
| ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
| FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
| details.
| 
| You should have received a copy of the GNU General Public License along with
| GNU Smalltalk; see the file COPYING.  If not, write to the Free Software
| Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  
|
 ======================================================================
 

|     Change Log
| ============================================================================
| Author        Date            Change
| bdiller       18 Oct 93       Corrrections and extensions to
|                               the kernel classes.
|
"

! Object methodsFor: 'debugging' !

inspect
"Open an Inspector window on self"
    Inspector new openOn: self
!!

! Dictionary methodsFor: 'debugging' !

inspect
"Open a DictionaryInspector window on self"
    DictionaryInspector new openOn: self
!!

! OrderedCollection methodsFor: 'enumerating the elements of a collection' !

allDo: aBlock atomize: op
"Recursively evaluate a block, aBlock, over successive elements of a 
collection.  The block is applied to collections from the bottom up.  The 
collection is recursively traversed until bottom level or non-collection 
elements are reached.  During the recursive ascent, the block is applied to 
each non-collection element, and then to its collection, the collection which 
contains the collection, and so on.  

This method is used to recursively manage the widgets stored in a view 
hierarchy from the bottom up (cf. TopLevelShells.display)"
    | subElements |
    1 to: self size do:
	[ :i | subElements := (self at: i) perform: op. subElements isNil 
					   ifTrue: [ aBlock value: (self at: i)] 
					   ifFalse: [
						     (subElements asOrderedCollection) allDo: aBlock atomize: op. 
						     aBlock value: (self at: i)]].
!!

! Symbol methodsFor: 'misc' !

numArgs
"Return number of selector arguments in a messsage selector symbol.
Number of selector arguments = number of  :'s in a symbol"
    | count |
    count := 0.
    self do: [ :value | value = $: ifTrue: [ count := count + 1 ]].
    ^count.
!!

!Class methodsFor: 'accessing instances and variables'!

"Corrections to methods for accessing class and pool variables.
Original implementation incorrectly assumed entire set of pool and class 
variables is represented in the class's instance variables, classVariables and 
sharedPools.  This version retrieves the set of pool and class variables by 
traversing up the the class hierarchy of the class"

classVarNames
"Return a class's class variable names.  Class variables are stored as 
associations.  This routine retrieves the class variable names from each 
association's key"
    | s |
    s _ Set new.
    classVariables notNil ifTrue: [
				   classVariables keysDo: [ :key | s add: key ]
				       ].
    ^s
!

allClassVarNames
"Return complete set of class variable names.  Traverse up the class's 
hierarchy to retrieve all the names"
    | superVarNames |
    superVarNames :=  self classVarNames.
    self allSuperclasses do: [ :each |
				      superVarNames addAll: each classVarNames].
    ^superVarNames.  
!

sharedPools
"Return class's shared pool dictionaries"
    | s |
    s _ Set new.
    sharedPools notNil ifTrue: [
				sharedPools do:
				    [ :sp | s add: (Smalltalk keyAtValue: sp) ]].
    ^s
!

allSharedPools
"Return complete set of pools.  Traverse up the class's hierarchy to retrieve 
all the pool dictionaries"
    | superPools |
    superPools :=  self sharedPools.
    self allSuperclasses do: [ :each |
				      superPools addAll: each sharedPools].
    ^superPools.  
!!

!ClassDescription methodsFor: 'accessing instances and variables'!

allInstVarNames
"p. 276 of the PurpleBook states that allInstanceVarNames should always return
an array of the receiver's instance variables"
    ^instanceVariables isNil ifTrue: [#()] ifFalse: [instanceVariables].
!!

!ClassDescription methodsFor: 'private'!

fileOutHeaderOn: aFileStream
"Write date and time stamp to aFileStream"
    | now |
    aFileStream nextPutAll: '''Filed out from ';
	nextPutAll: Version;
	nextPutAll: ' on '.
    now _ Date dateAndTimeNow.
    aFileStream print: (now at: 1);
	nextPutAll:  '  ';
	print: (now at: 2);
	nextPutAll: ' GMT''!'; nl; nl.
!

emitCategoryDescription: category toStream: aFileStream
"Write all methods belonging to 'category' to the file stream, aFileStream"
        aFileStream nextPutAll: '!';
        print: self;
        nextPutAll: ' methodsFor: '; nextPut: $';
        nextPutAll: category; nextPut: $';
        nextPutAll: '!'; nl.
!

fileOutDeclarationOn: aFileStream
"File out class definition"
        | aSet printFormattedSet |
        printFormattedSet := [ 
                aFileStream nextPutAll: ''''. 
                (aSet isEmpty) not ifTrue: [
                        aSet do: [ 
                                :element | aFileStream nextPutAll: element. 
                                aFileStream space]. 
                        aFileStream nextPutAll: ''''.
                        ].
        ].
        aFileStream nextPutAll: self superclass printString; space;
                nextPutAll: self kindOfSubclass; space;
                nextPutAll: '#';
                nextPutAll: name; nl; tab.

        aFileStream nextPutAll: 'instanceVariableNames: '.
        ((aSet := self instVarNames ) isEmpty)
                ifTrue:[aFileStream nextPutAll: '''''']
        ifFalse: printFormattedSet.
        aFileStream nl; tab.

    aFileStream
        nextPutAll: 'classVariableNames: '.
        (aSet := self classVarNames) isEmpty
        ifTrue:[aFileStream nextPutAll: '''''']
                ifFalse: printFormattedSet.
        aFileStream nl; tab.    

        aFileStream nextPutAll: 'poolDictionaries: '.
    (aSet := self sharedPools) isEmpty
        ifTrue:[aFileStream nextPutAll: '''''']
                ifFalse: printFormattedSet.
        aFileStream nl; tab.    

        aFileStream
                nextPutAll: 'category: '.
                category notNil ifTrue: [aFileStream nextPutAll: '''';
                        print: category; nextPutAll: '''']
                        ifFalse: [aFileStream print: category].
                aFileStream nl.
!

fileOutOn: aFileStream
"File out complete class description:  class definition, class and instance
methods"
    | categories theClass |
        name notNil 
         ifTrue: [
                self fileOutHeaderOn: aFileStream. 
                self fileOutDeclarationOn: aFileStream.
                aFileStream nextPut: $!; nl; nl
         ].

    categories _ Set new.
    methodDictionary isNil ifTrue: [ ^self ].
    methodDictionary do:
        [ :method | categories add: (method methodCategory) ].
        categories asSortedCollection do:
                [ :category | self emitCategory: category toStream: aFileStream ]
!!

!ClassDescription methodsFor: 'filing'!

fileOut: fileName
"Open a file selection dialog in the view.  File out complete class description
for the selected file"
   | aFileStream  |
    aFileStream _ FileStream open: fileName mode: 'w' ifFail: [ ^self error: 'Failed to open ''', fileName, '''' ].
         ('Filed out class: ', fileName) asSymbol printNl.
    name notNil
                ifTrue: [self fileOutOn: aFileStream. 
                        aFileStream nl;
                        nl.
                        self class fileOutOn: aFileStream]
                ifFalse: [self instanceClass fileOutOn: aFileStream. 
                        aFileStream nl;
                        nl.
                        self fileOutOn: aFileStream].
    aFileStream close.
!


fileOutCategory: categoryName to: fileName
"Prompt user for file name, file out all the methods belonging to the
method category, categoryName, to selected file"
    | aFileStream |
    aFileStream _ FileStream open: fileName mode: 'w'  ifFail: [ ^self error: 'Failed to open ''', fileName, '''' ].
    ('Filed out category: ', fileName) asSymbol printNl.
        self fileOutHeaderOn: aFileStream.
    self emitCategory: categoryName toStream: aFileStream.
    aFileStream close
!
        
fileOutSelector: selector to: fileName
"Prompt user for file name.  File out selector description to selected file"
    | aFileStream |
        aFileStream _ FileStream open: fileName mode: 'w' ifFail: [ ^self error: 'Failed to open ''', fileName, '''' ].
   ('Filed out selector: ', fileName) asSymbol printNl.
        self fileOutHeaderOn: aFileStream.
        self emitCategoryDescription: 
                (self compiledMethodAt: selector) methodCategory toStream: aFileStream.
        aFileStream nextPutAll: (self sourceCodeAt: selector);
        nl; nextPutAll: '!!'; close
!!

!ClassDescription methodsFor: 'private'!

emitCategory: category toStream: aFileStream
    "I write legal Smalltalk load syntax definitions of all of my methods
     are in the 'category' category to the aFileStream (sbb 91)"
    aFileStream nextPutAll: '!';
        print: self;
        nextPutAll: ' methodsFor: ''';
        nextPutAll: category;
        nextPutAll: '''!'.
    self selectors notNil 
        ifTrue: [ self selectors asSortedCollection do:
                      [ :selector | (self compiledMethodAt: selector) 
                                        methodCategory = category
                                        ifTrue: [ aFileStream nextPutAll: '

'                                           ;
                                                nextPutAll: (self sourceCodeAt: selector);
                                                nextPutAll: '!' ] ] ].
    aFileStream nextPutAll: '!

'
!!

! ClassDescription methodsFor: 'organization of messages and classes'!

"Fixed elementary error"

removeCategory: aString
    | selector method category |
    methodDictionary isNil
        ifTrue: [ ^self ].
    category _ aString asSymbol.
    methodDictionary associationsDo:
        [ :assoc | method _ assoc value.
                   method methodCategory = category
                        ifTrue: [ methodDictionary removeAssociation: assoc ] ].
                        
!!

!ClassDescription methodsFor: 'accessing'!

getName
"Method created to support browser.  Construct a name for class's Metaclass.
Note that Metaclasses have no name.  We did not want to override the name 
method in Metaclass because this will have side effects.  Other Smalltalk 
methods use this to determine whether the class is a Metaclass"
        name isNil 
                ifTrue: [^(self instanceClass name , ' class')].
        ^name.
!!

!Behavior methodsFor: 'accessing the methodDictionary'!
selectorAt: method
"Return selector key"
        ^methodDictionary keyAtValue: method.
!!


!Class methodsFor: 'instance creation'!

"Corrections for word and byte class instance creation.  Superclasses cannot
have any instance variables"

variableWordSubclass: classNameString
  instanceVariableNames: stringInstVarNames
  classVariableNames: stringOfClassVarNames
  poolDictionaries: stringOfPoolNames
  category: categoryNameString
    | meta |
    KernelInitialized ifFalse: [ ^nil ].
    self isVariable & ((self isPointers & (self instSize > 0)) | self isBytes) 
        ifTrue: [ ^self error: 'cannot create a word subclass from a non-word variable parent class' ].
        (TokenStream on: stringInstVarNames) contents isEmpty
                ifFalse: [^self error:'Instance variables not allowed' ].    
    meta _ self metaclassFor: classNameString.
    ^meta name: classNameString
            environment: Smalltalk
            subclassOf: self
            instanceVariableNames: stringInstVarNames
            variable: true
            words: true
            pointers: false
            classVariableNames: stringOfClassVarNames
            poolDictionaries: stringOfPoolNames
            category: categoryNameString
            comment: nil
            changed: nil
!

variableByteSubclass: classNameString
  instanceVariableNames: stringInstVarNames
  classVariableNames: stringOfClassVarNames
  poolDictionaries: stringOfPoolNames
  category: categoryNameString
    | meta |
    KernelInitialized ifFalse: [ ^nil ].
    self isVariable & ((self isPointers & (self instSize > 0)) | self isWords)
        ifTrue: [ ^self error: 'Cannot create variable byte subclass from non-byte parent class' ].
        (TokenStream on: stringInstVarNames) contents isEmpty
                ifFalse: [^self error: 'Instance variables not allowed' ].    
        meta _ self metaclassFor: classNameString.
    ^meta name: classNameString
            environment: Smalltalk
            subclassOf: self
            instanceVariableNames: stringInstVarNames
            variable: true
            words: false
            pointers: false
            classVariableNames: stringOfClassVarNames
            poolDictionaries: stringOfPoolNames
            category: categoryNameString
            comment: nil
            changed: nil
!!


!Metaclass methodsFor: 'basic'!

"Extensions and modifications made to support the ability to make changes to
existing classes"

name: newName
    environment: aSystemDictionary
    subclassOf: superclass
    instanceVariableNames: stringOfInstVarNames
    variable: variableBoolean
    words: wordBoolean
    pointers: pointerBoolean
    classVariableNames: stringOfClassVarNames
    poolDictionaries: stringOfPoolNames
    category: categoryName
    comment: commentString
    changed: changed
    | aClass variableString variableArray sharedPoolNames poolName pool 
      className classVarDict  needToRecompileMetaclasses needToRecompileClasses |

    className _ newName asSymbol.
    aClass _ aSystemDictionary at: className ifAbsent: [ nil ].
    aClass isNil
        ifTrue: [ ^self newMeta: newName
                        environment: aSystemDictionary
                        subclassOf: superclass
                        instanceVariableNames: stringOfInstVarNames
                        variable: variableBoolean
                        words: wordBoolean
                        pointers: pointerBoolean
                        classVariableNames: stringOfClassVarNames
                        poolDictionaries: stringOfPoolNames
                        category: categoryName
                        comment: commentString
                        changed: changed ].

        needToRecompileMetaclasses := false.

        (aClass isKindOf: Class)
                ifFalse: [^self error: aClass name, 'is not a Class'].
                                
        (aClass superclass ~~ superclass) 
                        ifTrue: [^self error: 'Cannot change superclass of ' , aClass name].    

        variableBoolean 
"If the class is changed to a variable word or byte subclass,
then check to see whether all variable subclasses are the same"
                ifTrue:[ ((aClass isWords ~~ wordBoolean )
                | (aClass isPointers ~~ pointerBoolean))
                & (pointerBoolean not)
                ifTrue: [aClass allSubclassesDo: [:sc | 
                (sc isVariable) ifTrue: [
                        (sc isWords ~~ wordBoolean) 
                                ifTrue: [^self error: 'cannot change to a variable word 
                                subclass because class has variable non-word subclasses']. 
                        (sc isPointers ~~ pointerBoolean) 
                                ifTrue: [^self error: 'cannot change class to a variable byte 
                                class because class has variable non-byte subclasses' ].
                ]]]].

        superclass notNil 
                ifTrue: [ "Inherit instance variables from parent"
                                        variableString _ superclass instanceVariableString
                                ]
                ifFalse: [ variableString _ '' ].
    variableString _ variableString , stringOfInstVarNames.
    variableArray _ self parseVariableString: variableString.
    1 to: variableArray size do:
        [ :i | variableArray at: i put: (variableArray at: i) asSymbol ].

"If instance or indexed variables change, update 
instance variables and instance spec of the class and all its subclasses "
        (needToRecompileClasses := variableArray ~= aClass allInstVarNames) 
        | (aClass isVariable ~~ variableBoolean)
        | (aClass isWords ~~ wordBoolean )
        | (aClass isPointers ~~ pointerBoolean)

                ifTrue:[aClass updateInstanceVars: variableArray 
                variable: variableBoolean words: wordBoolean pointers: pointerBoolean].

        classVarDict _ (self parseToDict: stringOfClassVarNames).
        
        needToRecompileMetaclasses := false.
                
    classVarDict _ (self parseToDict: stringOfClassVarNames).

   aClass classPool isNil 
                ifTrue: [ aClass setClassVariables: classVarDict ]
                ifFalse: [ classVarDict associationsDo:
                                        [ :assoc | (aClass classPool includesKey: assoc key)
                                                        ifFalse: 
                                                        [ aClass addClassVarName: 
                                                        assoc key ]]. 
                                                        
                                aClass classPool keysDo: 
                                        [ :aKey | (classVarDict includesKey: aKey)
                                                ifFalse: 
                                                        [aClass removeClassVarName: aKey.
                                                        needToRecompileMetaclasses := true.]].  
                                
                                ].
 
    sharedPoolNames _ self parseVariableString: stringOfPoolNames.
    1 to: sharedPoolNames size do:
        [ :i | poolName _ (sharedPoolNames at: i) asSymbol.
                        (poolName at: 1) isUppercase 
                                ifFalse: [^self error: 'Pool name must begin with an uppercase 
                                letter'].

               "??? Should this create the pool if not there? (sbb 91)"
               pool _ aSystemDictionary
                        at: poolName
                        ifAbsent: [ ^self error: 'Pool name ', poolName ,
                                                 ' does not exist' ].
              sharedPoolNames at: i put: pool ].
                          
    aClass sharedPools isNil 
                ifTrue: [ aClass setSharedPools: sharedPoolNames ]
                ifFalse: [ sharedPoolNames do:
                                        [ :dict | (aClass sharedPools includes: dict)
                                                        ifFalse: 
                                                        [ aClass addSharedPool: 
                                                        dict ] ]. 
                                                        
                                aClass sharedPools do: 
                                        [ :dict | (sharedPoolNames includes: dict)
                                                ifFalse: 
                                                        [aClass removeSharedPool: dict.
                                                        needToRecompileMetaclasses := true.]].  
                                
                                ].

    "### not done??"
    aClass category: categoryName. "### need to remove the old category maybe (sbb 91)"
        "from where??? (bdiller)"

"Please note that I need to recompile the classes in this sequence;
otherwise, the same error is propagated to each selector which is compiled
after an error is detected even though there are no further compilation 
errors.

Apparently, there is a bug in the primitive compileString.  This can be cleaned
up later"

         
        (needToRecompileClasses | needToRecompileMetaclasses)
                ifTrue: [ stdout nextPutAll: 'Recompiling Classes...'; nl.
                                needToRecompileClasses :=  aClass compileAll.
                                (needToRecompileMetaclasses)
                                        ifTrue: [aClass class compileAll].
                                needToRecompileClasses :=  aClass compileAllSubclasses.
                                (needToRecompileMetaclasses)
                                        ifTrue: [aClass class compileAllSubclasses].
                                ].      
        ^aClass
!!

!ClassDescription methodsFor: 'private'!

updateInstanceVars: variableArray variable: variableBoolean words: wordBoolean 
pointers: pointerBoolean
"Update instance variables and instance spec of the class and all its subclasses"
        | instVarMap startOfInstanceVars endOfInstanceVars newInstanceVars 
        oldInstVars map iv instances cvtObject |
         
        Smalltalk compact.

        startOfInstanceVars := self superclass instSize + 1.
        endOfInstanceVars :=  self instSize.
        newInstanceVars := variableArray copyFrom: startOfInstanceVars 
                to: variableArray size. 
        oldInstVars :=  self allInstVarNames. 
        instVarMap := Array new: newInstanceVars size.
        startOfInstanceVars to: endOfInstanceVars do: [ :i | 
                map := newInstanceVars indexOf: (oldInstVars  at: i). 
                map > 0 ifTrue: [instVarMap at: map put: i]
                ].

"Fix up all subclasses.  Ignore the metas"
        self allSubclassesDo: [ :sc | (sc isKindOf: Metaclass) not ifTrue:[
                        iv := sc allInstVarNames copyReplaceFrom: startOfInstanceVars 
                        to: endOfInstanceVars with: newInstanceVars.
                        sc setInstanceVariables: iv.
"If a class is changed from a non-variable to variable class, update all its 
non-variable subclasses to be the same.  Update size information of class's
instanceSpec"
                        ((self isVariable ~~ variableBoolean) & variableBoolean & 
                                sc isVariable not)
                                ifTrue: [ sc setInstanceSpec: variableBoolean words: wordBoolean
                                        pointers: pointerBoolean instVars: sc allInstVarNames size]
                                ifFalse: [ sc setInstanceSpec: sc isVariable words: sc isWords
                                        pointers: sc isPointers instVars: sc allInstVarNames size].

"Mutate all instances of the class to conform to new memory model of the class.

Warning: save off instances in a temp first rather than directly iterate 
through all the object's instances using allInstancesDo:.  The do: block will 
manufacture temporary duplicates of the object to mutate the instances to the 
new memory model.  As a result, allInstancesDo: will also recursively iterate 
over these temporary duplicates"

                instances := sc allInstances.
                instances do: [ :each | each mutate: instVarMap startAt: 
                        startOfInstanceVars instSizeDiff: (oldInstVars size  - 
                        variableArray size) cvtObject: true ].
                ]].

"Now update class instance vars "       
    self setInstanceVariables: variableArray.

"If indexed representation changes, do not convert"
        cvtObject := ((variableBoolean & self isVariable) & 
                (self isWords ~~ wordBoolean ) | 
                        (self isPointers ~~ pointerBoolean)) not.

    self setInstanceSpec: variableBoolean words: wordBoolean
        pointers: pointerBoolean instVars: variableArray size.

"Need to pass old instance spec to this invocation to make sure that index 
variable conversions are handled correctly.  Maybe only support byte <-> word 
conversions, or give a special warning that such conversions will change 
representation.  For the time being, pass; the author does not think you should 
convert in these cases, because any conversion will change the original data"
        instances := self allInstances.
        instances do: [ :each | each mutate: instVarMap startAt: startOfInstanceVars
                        instSizeDiff: (oldInstVars size  - variableArray size) 
                        cvtObject: cvtObject].
!!

!Object methodsFor: 'private'!

mutate: instVarMap startAt: start instSizeDiff: adjustment cvtObject: cvtObject
"Mutate object to new class representation"
    | class aCopy mappedValue end numIndexedVariables  |
    class _ self class.
    class isVariable
"Allocate a new object based on the new class size derived from basicSize - adjustment"
        ifTrue: [ aCopy _ class basicNew: (numIndexedVariables := 
                        (self basicSize - adjustment))]
                ifFalse: [ aCopy _ class basicNew ].

        cvtObject ifFalse: [^self become: aCopy].

        end := instVarMap size + start - 1.
        
"Copy the instance variables, if any"
    1 to: start - 1 do:
        [ :i | aCopy instVarAt: i
                     put: (self instVarAt: i) ].
                                 
"Copy old instance variables to their new positions using instVarMap"
    start to: end do:
        [ :i | mappedValue := instVarMap at: i - start + 1. 
                mappedValue notNil 
                        ifTrue: [aCopy instVarAt: i
                                                put: (self instVarAt: mappedValue)]].

        end + 1 to: class instSize do:
        [ :i | aCopy instVarAt: i 
                                put: (self instVarAt: i + adjustment) ].

"Copy the indexed variables, if any.

Note that conversions to different kind of indexed variables is not handled 
correctly.  This should probably be disallowed because the representation is 
entirely different.  We really need to have the old object in order to use its 
old instanceSpec"

    class isVariable
        ifTrue: [ 1 + adjustment to: numIndexedVariables + adjustment do:
                    [ :i | aCopy basicAt: i - adjustment
                                 put: (self specialBasicAt: i).
                                                ]].
        self become: aCopy.
!!

!Metaclass methodsFor: 'basic'!

"Removed comments which no longer apply now that complete Metaclass support for 
changing existing classes has been provided"

newMeta: newName
    environment: aSystemDictionary
    subclassOf: superclass
    instanceVariableNames: stringOfInstVarNames
    variable: variableBoolean
    words: wordBoolean
    pointers: pointerBoolean
    classVariableNames: stringOfClassVarNames
    poolDictionaries: stringOfPoolNames
    category: categoryName
    comment: commentString
    changed: changed
    | aClass variableString variableArray sharedPoolNames poolName pool |

    sharedPoolNames _ self parseVariableString: stringOfPoolNames.
    1 to: sharedPoolNames size do:
        [ :i | poolName _ (sharedPoolNames at: i) asSymbol.
               (poolName at: 1) isUppercase
                   ifFalse: [ ^self error: 'Pool name ', poolName,
                                  ' does not begin with an uppercase letter' ].
               pool _ aSystemDictionary at: poolName
                        ifAbsent: [ ^self error: 'Pool name ', poolName ,
                                        ' does not exist' ].
              sharedPoolNames at: i put: pool ].
    aClass _ self new.
    instanceClass _ aClass.
    aSystemDictionary at: (newName asSymbol) put: aClass.
    aClass superclass: superclass.
    aClass setName: newName asSymbol.
    superclass notNil
        ifTrue: [ superclass addSubclass: aClass.
                  "Inherit instance variables from parent"
                  variableString _ superclass instanceVariableString
                  ]
        ifFalse: [ variableString _ '' ].
    variableString _ variableString , stringOfInstVarNames.
    variableArray _ self parseVariableString: variableString.
    1 to: variableArray size do:
        [ :i | variableArray at: i put: (variableArray at: i) asSymbol ].
    aClass setInstanceVariables: variableArray.
    aClass setInstanceSpec: variableBoolean words: wordBoolean
        pointers: pointerBoolean instVars: variableArray size.
    aClass setClassVariables: (self parseToDict: stringOfClassVarNames).
    aClass setSharedPools: sharedPoolNames.
    aClass category: categoryName.
    aClass comment: commentString.
    ^aClass
!!

!Behavior methodsFor: 'evaluating'!


evaluate: code
"Evaluate Smalltalk expression in 'code' and return result.  Some modifications
to the Smalltalk grammar were required to make this work (cf. our changes in 
mstcomp.c and mst.y)"

        | result |
    (code isKindOf: PositionableStream)
        ifTrue: [ code _ code contents ].
    (code isMemberOf: String)
        ifFalse: [ code _ code asString ].
"Manually append a '!' to force parser execute code"
        result := self evalString: ( code , '!').
        ^result.
!


evaluate: code ifError: block
 "Evaluate 'code'.  If a parsing error is detected, invoke 'block'"
        | result |
   (code isKindOf: PositionableStream)
        ifTrue: [ code _ code contents ].
    (code isMemberOf: String)
        ifFalse: [ code _ code asString ].
        result := self evalString: ( code , '!') ifError: block.
        ^result.
!

evaluate: code to: anObject ifError: block
"Evaluate Smalltalk expression as part of anObject's method definition.  This 
method is used to support Inspector expression evaluation.  If a parsing error 
is encountered, invoke error block, 'block'" 
        | result |
   (code isKindOf: PositionableStream)
        ifTrue: [ code _ code contents ].
    (code isMemberOf: String)
        ifFalse: [ code _ code asString ].
        result := self evalString: ( code , '!') to: anObject ifError: block.
        ^result.
!

evaluate: code to: anObject 
"Evaluate Smalltalk expression as part of anObject's method definition"
        | result |
   (code isKindOf: PositionableStream)
        ifTrue: [ code _ code contents ].
    (code isMemberOf: String)
        ifFalse: [ code _ code asString ].
        result := self evalString: ( code , '!') to: anObject.
        ^result.
!

evaluate: code notifying: requestor
"Evaluate Smalltalk expression in 'code'. If a parsing error is encountered, 
invoke error block, 'block'"
    | method aStream |
        ^self evaluate: code ifError: [:fname :lineNo :errorString | 
                requestor error: ('line ', lineNo printString, ': ' , errorString) ].
!!

!Behavior methodsFor: 'testing the method dictionary'!
getByteCodeForSpecialSelector: selectorName
"If selectorName is a special-cased selector, return its byte code.  Otherwise, 0 is returned"
    <primitive: 262>
    ^self primitiveFailed
!


whichSelectorsReferToByteCode: aByteCode
"Return the collection of selectors in the class which reference the byte code, aByteCode"
    | s method specialSelector |
    s _ Set new.
    methodDictionary isNil
        ifTrue: [ ^s ].
    methodDictionary associationsDo:
        [ :assoc |  method _ assoc value.
                                ((1 to: method numBytecodes) detect:
                                [ :i | aByteCode = (method bytecodeAt: i)]
                                ifNone: [0]) > 0 
                                        ifTrue: [ s add: assoc key ].
        ].
    ^s
!!

!ClassDescription methodsFor: 'compiling'!

compile: code classified: categoryName
"If the method, 'code', successfully compiles, install method in method category, categoryName.  Return the 
method"
    | method |
    method _ self compile: code.
    method notNil 
                ifTrue: [method methodCategory: categoryName].
    ^method
!

compile: code classified: categoryName notifying: requestor
"Compile method source and install in method category, categoryName.  If there 
are parsing errors, invoke requestor's exception handler.  Return the method"
    | method |
    method _ self compile: code notifying: requestor.
    method notNil 
                ifTrue: [method methodCategory: categoryName].
    ^method
!

compile: code classified: categoryName ifError: block
"Compile method source and install in method category, categoryName.  If there 
are parsing errors, invoke exception block, 'block'.  Return the method"
    | method |
    method _ self compile: code ifError: block.
    method notNil 
                ifTrue: [method methodCategory: categoryName].
    ^method
!!

!Behavior methodsFor: 'creating method dictionary'!

compile: code ifError: block
"Compile method source.  If there are parsing errors, invoke exception block, 'block'.
Return result of compilation"
    (code isKindOf: PositionableStream)
        ifTrue: [ code _ code contents ].
    (code isMemberOf: String)
        ifFalse: [ code _ code asString ].
    ^self compileString: code ifError: block.
!!

! Object methodsFor: 'built ins' !
specialBasicAt: index
"Like basicAt: but without bounds checking.  This method is used to support 
instance mutation when an instance's class definition is changed"
    <primitive: 263>
!!

!Behavior methodsFor: 'built ins'!

"Primitives added to support expression evaluation"

evalString: aString
    <primitive: 256>
    ^self primitiveFailed
!

evalString: aString ifError: aBlock
    <primitive: 257>
    ^self primitiveFailed
!

evalString: aString to: anObject
    <primitive: 258>
    ^self primitiveFailed
!

evalString: aString to: anObject ifError: aBlock
    <primitive: 259>
    ^self primitiveFailed
!!

!SystemDictionary methodsFor: 'builtins'!

compact
"Force a full garbage collection"
    <primitive: 260>
    ^self primitiveFailed
!

halt
"Interrupt interpreter"
    <primitive: 261>
    ^self primitiveFailed
!!

!Behavior methodsFor: 'creating method dictionary'!
"Enhancements to support browser system.  Log compilations to output window to 
facilitate error identification"

compileAll
        ('Recompiling class: ' , self getName, '...') asSymbol printNl.
    methodDictionary notNil
        ifTrue: [ methodDictionary keysDo: 
                      [ :selector | self recompile: selector ] ]
!

compileAll: aNotifier
        ('Recompiling class: ' , self getName, '...') asSymbol printNl.
    methodDictionary notNil
        ifTrue: [ methodDictionary keysDo: 
                      [ :selector | self recompile: selector notifying: aNotifier] ]
!

compileAllSubclasses: aNotifier
    self allSubclassesDo: [ :subclass | subclass compileAll: aNotifier ]
!

recompile: selector
        ('Recompiling selector: ' , selector) asSymbol printNl.
    self compile: (self sourceCodeAt: selector) classified: 
                (self compiledMethodAt: selector) methodCategory.
!

recompile: selector notifying: aNotifier
        ('Recompiling selector: ' , selector) asSymbol printNl.
        self compile: (self sourceCodeAt: selector) classified: 
                ((self compiledMethodAt: selector) methodCategory) notifying: aNotifier.
!!

"Corrections to handle recursive dictionaries such as the Smalltalk dictionary
where Smalltalk is a member of itself"

!Dictionary methodsFor: 'printing'!
printOn: aStream
"Print the value of each dictionary entry to aStream.  If the 
dictionary contains itself as an entry, do not recursively call this routine"
    aStream nextPutAll: self class name , ' (' .
    self associationsDo:
        [ :assoc | (assoc value ~~ self) ifTrue: [aStream print: assoc;
                       nextPut: Character space ] ifFalse: [aStream print: assoc key; 
                           nextPutAll: '->'; print: assoc key; nextPut: Character space ]].
    aStream nextPut: $)
!!

!Dictionary methodsFor: 'storing'!

storeOn: aStream
"Store each dictionary entry to aStream.  If the dictionary contains itself as an entry, 
do not recursively call this routine 
"
    | hasElements |
    aStream nextPutAll: '(', self class name , ' new'.
    hasElements _ false.
    self associationsDo:
        [ :assoc | aStream nextPutAll: ' at: ';
                       store: assoc key;
                       nextPutAll: ' put: '.
                        (assoc value ~~ self) 
                                ifTrue: [aStream store: assoc value]
                                ifFalse: [aStream store: assoc key].
                       aStream nextPut: $;.
                   hasElements _ true ].
    hasElements ifTrue: [ aStream nextPutAll: ' yourself' ].
    aStream nextPut: $)
!!

!Dictionary methodsFor: 'misc math methods'!

= aDictionary
"Test whether aDictionary is identical to iself by comparing each element.  If
the dictionary contains itself as an entry, do not recursively call this 
routine"
    self class == aDictionary class
        ifFalse: [ ^false ].
    tally ~= aDictionary size ifTrue: [ ^false ].
    self associationsDo:
"Do not recurse on recursive set"
        [ :assoc | (assoc value ~~ self) ifTrue: [ 
                        assoc value ~= (aDictionary at: assoc key ifAbsent: [ ^false ])
                        ifTrue: [ ^false ] 
                        ]
                ].
    ^true
!!

"bdiller Wed Nov  3 12:43:17 PST 1993

fixed rehash bug -- use findKeyIndex: rather than hash function
directly to find next element that hashes to this index.

The original version failed in the following case:

There is a contiguous block of table entries prior to the element
deleted in the hash table.  There is no element in the contiguous
block of entries following the deleted element which hashes to deleted
index and there is at least one element, q, in this block which
follows the element which hashes directly to an entry in the prior
block.

In the original algorithm, the deleted entry will not be filled by the
element q.  As a result, invoking findKeyIndex: on q and any element
which hashes to an entry in the prior block will come up empty.

Solution: use findKeyIndex: rather than the hash function to restore
table's integrity.

Loop until hash table is rendered consistent.  There is no reason to
test that the number of iterations is less than or equal to size of
table.  The loop will naturally terminate after a finite number of
iterations.  A contiguous block of entries which follows the deleted
element must be finite sized since the table is itself finite sized."

!Dictionary methodsFor: 'private methods'!
rehashObjectsAfter: index
    "Rehashes all the objects in the collection after index to see if any of
    them hash to index.  If so, that object is copied to index, and the
    process repeats with that object's index, until a nil is encountered."

    | i size count assoc |
    i _ index.
    size _ self basicSize.
        [ i _ i \\ size + 1.
                assoc _ self basicAt: i.
                assoc notNil]
        whileTrue: [
                index := self findKeyIndex: assoc key.
                (self basicAt: index) isNil
                ifTrue: [ self basicAt: index put: assoc.
                                self basicAt: i put: nil.  "Be tidy"]].

!!

!IdentityDictionary methodsFor: 'private methods'!
rehashObjectsAfter: index
    "Rehashes all the objects in the collection after index to see if any of
    them hash to index.  If so, that object is copied to index, and the
    process repeats with that object's index, until a nil is encountered."

    | i size count key |
    i _ index.
    size _ self basicSize.
        [ i _ i \\ size + 1.
                key _ self basicAt: i.
                key notNil]
        whileTrue: [
                index := self findKeyIndex: key.
                (self basicAt: index) isNil
                ifTrue: [ self basicAt: index put: key.
                                values basicAt: index put: (values basicAt: i).
                                self basicAt: i put: nil.  "Be tidy"
                        values basicAt: i put: nil."Be tidy"]].

!!

! OrderedCollection methodsFor: 'removing'!

remove: element
        self removeAtIndex: (self indexOf: element).
!

removeAtIndex: index
"Remove entry at 'index'.  All the collection entries which follow element 
index are pushed down one entry.  The last entry is filled with nil.  The 
instance variable, lastIndex, is decremented"
    index _ index + firstIndex - 1.
    (index >= firstIndex and: [ index <= lastIndex ])
        ifFalse: [^self error: 'index out of bounds for ordered collection'].
    index  to: lastIndex - 1 do:
        [ :i |          
                self basicAt: i  put: (self basicAt: i + 1)
        ].
        self basicAt: lastIndex  put: nil.
        lastIndex := lastIndex - 1
!!

"questionable change -- use identity test to improve efficiency"
"
!Dictionary methodsFor: 'accessing'!

keyAtValue: value ifAbsent: exceptionBlock
    self associationsDo:
        [ :assoc | value == assoc value
                     ifTrue: [ ^assoc key ] ].
    ^exceptionBlock value
! !
"
!OrderedCollection methodsFor: 'adding'!

add: newObject afterIndex: index
    index _ index + firstIndex - 1.
    (index >= firstIndex and: [ index <= lastIndex ])
        ifTrue: [ ^self at: index + 1 insertObject: newObject ]
        ifFalse: [ self error: 'index out of bounds for ordered collection' ]
!

add: newObject beforeIndex: index
    index _ index + firstIndex - 1.
    (index >= firstIndex and: [ index <= lastIndex ])
        ifTrue: [ ^self at: index  insertObject: newObject ]
        ifFalse: [ self error: 'index out of bounds for ordered collection' ]
! !

!OrderedCollection methodsFor: 'private methods'!

at: index insertObject: anObject
    lastIndex = self basicSize ifTrue: [ self growLast ].
                lastIndex _ lastIndex + 1.
    lastIndex to: index by: -1 do:
        [ :i | self basicAt: i + 1
                    put: (self basicAt: i) ].
    self basicAt: index put: anObject.
    ^anObject
! !

!SystemDictionary methodsFor: 'builtins'!
enableGC: aBoolean
"Enable GC of Smalltalk objects referenced in callins/callouts"
    <primitive: 264>
    ^self primitiveFailed
!!

