(* Copyright (C) 1992, Digital Equipment Corporation           *)
(* All rights reserved.                                        *)
(* See the file COPYRIGHT for a full description.              *)

(* Last modified on Thu May  5 14:00:34 PDT 1994 by kalsow     *)
(*      modified on Mon Apr 26 19:00:42 PDT 1993 by muller     *)

UNSAFE MODULE RTHeapMap;

IMPORT RT0, RTType, RTHeapRep, RTModule, RTMapOp;
IMPORT RTIO;

VAR DEBUG := FALSE;

TYPE
  Byte = RTMapOp.Byte;
  BP   = RTMapOp.BP;
  IP   = RTMapOp.IP;
  AP   = RTMapOp.AP;
  PC   = RTMapOp.PC;
  Op   = RTMapOp.T;

(*------------------------------------------------ walking of single refs ---*)

PROCEDURE WalkRef (h: ObjectPtr;  v: Visitor) =
  VAR
    tc: RT0.Typecode := h.typecode;
    t: RT0.TypeDefn;
  BEGIN
    IF (tc # RTHeapRep.Fill_1_type) AND (tc # RTHeapRep.Fill_N_type) THEN
      <* ASSERT tc # 0 *>
      t := RTType.Get (tc);
      IF DEBUG THEN
        Txt("Ref: tc="); Int(tc); Txt(" cell="); Adr(t); Txt(" ref=");
        Adr(h + ADRSIZE (h^)); Txt("\n");
      END;
      DoWalkRef (t, h+ADRSIZE (h^), v);
    END;
  END WalkRef;

PROCEDURE DoWalkRef (t: RT0.TypeDefn;  a: ADDRESS;  v: Visitor) =
  BEGIN
    IF (t.parent # NIL) THEN
      DoWalkRef (t.parent, a, v);
      INC (a, t.dataOffset);
    END;
    Walk (t.gc_map, a, v);
  END DoWalkRef;
  
(*------------------------------------------------ walking of the globals ---*)

TYPE ModuleMap = UNTRACED REF ARRAY OF RT0.ModulePtr;
VAR global_map: ModuleMap := NIL;

PROCEDURE WalkGlobals (v: Visitor) =
  VAR m: RT0.ModulePtr;
  BEGIN
    IF (global_map = NIL) THEN BuildGlobalMap () END;
    FOR i := 0 TO LAST (global_map^) DO
      m := global_map[i];
      IF DEBUG THEN Txt("Module: "); Int(i); Txt("  "); Adr(m); Txt("\n"); END;
      Walk (m.gc_map, m, v);
    END;
  END WalkGlobals;

PROCEDURE BuildGlobalMap () =
  VAR n: INTEGER;  m: RT0.ModulePtr;  max := RTModule.Count()-1;
  BEGIN
    (* first, count the modules that qualify *)
    n := 0;
    FOR i := 0 TO max DO
      m := RTModule.Get (i);
      IF (m # NIL) AND (m.gc_map # NIL) THEN INC (n); END;
    END;

    (* allocate the space *)
    global_map := NEW (ModuleMap, n);

    (* and fill it in *)
    n := 0;
    FOR i := 0 TO max DO
      m := RTModule.Get (i);
      IF (m # NIL) AND (m.gc_map # NIL) THEN
        global_map[n] := m;
        INC (n);
      END;
    END;
  END BuildGlobalMap;

PROCEDURE WalkModuleGlobals (v: Visitor;  mod: CARDINAL) =
  VAR m := RTModule.Get (mod);
  BEGIN
    IF (m # NIL) AND (m.gc_map # NIL) THEN
      Walk (m.gc_map, m, v);
    END;
  END WalkModuleGlobals;

(*------------------------------------------------------- internal walker ---*)

PROCEDURE Walk (map: ADDRESS;  val: ADDRESS;  v: Visitor) =
   VAR
     pc    : ADDRESS := map; (* program counter *)
     x     : ADDRESS := val; (* value pointer *)
     op    : Op;             (* current opcode *)
     n, m  : INTEGER;
     elts  : ADDRESS;
     stack : RTMapOp.Stack;
     z     : INTEGER;
  BEGIN
    IF (val = NIL) OR (pc = NIL) THEN RETURN END;
    IF DEBUG THEN
      Txt("Walk: map="); Adr(map); Txt(" val="); Adr(val);
      Txt(" visitor="); Adr(LOOPHOLE (v, ADDRESS)); Txt("\n");
    END;
    LOOP
      op := LOOPHOLE (pc, PC)^;
      IF DEBUG THEN
        Txt("  pc="); Adr(pc); Txt("  op="); Int(ORD (op)); Txt("\n");
      END;
      INC (pc, ADRSIZE (Op));
      CASE op OF
      | Op.Mark =>
          RTMapOp.Push (stack, pc, -1);

      | Op.Stop =>
          LOOP
            IF (stack.top <= 0) THEN RETURN END;
            (* otherwise, resume a suspended walk *)
            WITH tos = stack.data [stack.top-1] DO
              pc := tos.pc;
              DEC (tos.count);
              IF (tos.count >= 0) THEN EXIT END;
              DEC (stack.top);
            END;
          END;

      | Op.Array_1, Op.Array_2, Op.Array_3, Op.Array_4,
        Op.Array_5, Op.Array_6, Op.Array_7, Op.Array_8 =>
          z := RTMapOp.OpArgBytes [op];
          n := RTMapOp.GetInt (pc, z); (* number of elements *)
          WITH tos = stack.data [stack.top-1] DO
            IF tos.count = -1 THEN (* first time *) tos.count := n END;
            DEC (tos.count);
            IF (tos.count > 0) THEN  pc := tos.pc  ELSE  DEC (stack.top)  END;
          END;

      | Op.OpenArray_1, Op.OpenArray_2 =>
          z := RTMapOp.OpArgBytes [op];
          n := RTMapOp.GetInt (pc, z); (* open array depth *)
          elts := LOOPHOLE (x, AP)^;  (* array elements *)
          INC (x, ADRSIZE (ADDRESS));

          (* compute the total number of elements to visit *)
          m := 1;
          FOR j := 1 TO n DO
            m := m * LOOPHOLE (x, IP)^;
            INC (x, ADRSIZE (INTEGER));
          END;

          (* push the remainder of the map with the element count *)
          IF (m <= 0) THEN RETURN END;
          RTMapOp.Push (stack, pc, m-1);
          x := elts;
            
      | Op.Ref =>
          v (x);
          INC (x, ADRSIZE (ADDRESS));

      | Op.PushPtr =>
          RTMapOp.Push (stack, x + ADRSIZE (AP), -1);
          x := LOOPHOLE (x, AP)^;

      | Op.Return =>
          DEC (stack.top);
          x := stack.data [stack.top].pc;

      | Op.Skip_1, Op.Skip_2, Op.Skip_3, Op.Skip_4,
        Op.Skip_5, Op.Skip_6, Op.Skip_7, Op.Skip_8 =>
          INC (x, RTMapOp.OpSize [op]);

      | Op.SkipF_1 =>
          INC (x, LOOPHOLE (pc, BP)^);
          INC (pc, ADRSIZE (Byte));

      | Op.SkipF_2, Op.SkipF_3, Op.SkipF_4,
        Op.SkipF_5, Op.SkipF_6, Op.SkipF_7, Op.SkipF_8 =>
          z := RTMapOp.OpArgBytes [op];
          n := RTMapOp.GetInt (pc, z);
          INC (x, n);

      | Op.SkipB_1 =>
          DEC (x, LOOPHOLE (pc, BP)^);
          INC (pc, ADRSIZE (Byte));

      | Op.SkipB_2, Op.SkipB_3, Op.SkipB_4,
        Op.SkipB_5, Op.SkipB_6, Op.SkipB_7, Op.SkipB_8 =>
          z := RTMapOp.OpArgBytes [op];
          n := RTMapOp.GetInt (pc, z);
          DEC (x, n);

      ELSE
          <*ASSERT FALSE*>
      END;
    END;
  END Walk;

PROCEDURE Txt(t: TEXT) =
  BEGIN RTIO.PutText (RTIO.stderr, t) END Txt;
PROCEDURE Int(i: INTEGER) =
  BEGIN RTIO.PutInt  (RTIO.stderr, i) END Int;
PROCEDURE Adr(a: ADDRESS) =
  BEGIN RTIO.PutHexa (RTIO.stderr, LOOPHOLE (a, INTEGER)) END Adr;

BEGIN
END RTHeapMap.

